/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 July 2003
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>

#include <stdlib.h>
#include <math.h>
#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <ipport.h>
#include <sense.h>
#include <line.h>
#include <debug.h>

extern char outputfile[];

extern Scheduler * sched;
const int MAXCODEWORDSONLINE = 0;

Line::Line(double b, int s)
{
	bw = b ;
	spreading = s;
	numcodesonline= 0;
	q = new Queue(MAXCODEWORDSONLINE);
	fout = fopen(outputfile, "w");
	if (fout == NULL)
		dprintf("res.cc: Line::Line(): Could not open output file for writing\n");

	line = new int[(int)spreading];
	for(int count =(int)(spreading) - 1; count >=0; count--)
			line[count] = 0;

	drops = 0;
	receives =0;

	pktcnt =0;
	bytes=0;
}

Line::~Line()
{
}

//The Line message handling function: It queues the packet, checks if the port is busy, processes it if not.
int Line::process(Packet  *p)
{

//Bug fix: Add propagation delay to time delay

	double prop = 1000 / 2e8;
	double delay = (double)(p->size) *(8.0)/(bw/(double)spreading);	
	delay+=prop;
	q->enqueue(p);


	//Check if this code destroys any other code on the line. If so record.
	check(p);

	dprintf("Line: Scheduling packt delay: %.6f\n", delay);
	dprintf("Line: Number of packets on line : %d\n", q->qlength());

	sched->scheduleevent(this, p, delay);
}



//The Line event handling fucntion: Dequeues a packet, processes it, 
//checks if an input interface is waiting and schedules it if it is.
void Line::handleevent(Packet *p)
{
	
	dprintf("res.cc: Line::Line(): Packet departing : id: %d\n", p->id);
	q->dequeue(p);
	dprintf("Line: Number of packets on line : %d\n", q->qlength());

	pktcnt++;
	if(p->drop ==1)
	{
		drops++;
	}
	else
	{
		bytes+=p->size;
		receives++;
	}
	
	char state[20];
	if(p->drop == 1)
		strcpy(state, "D");
	else 
		strcpy(state, "R");

	double delay =0;
	if(p->drop != 1)
		delay = p->txtime - p->timestamp;
	else delay =0;
	fprintf(fout, "%.12f %d %d.%d %d.%d %d %d %d %s %.9f\n", p->txtime, p->size, p->srcaddr, p->srcport, p->dstaddr, p->dstport, p->id, p->codeword, p->defers, state, delay);

//Remove line entrys on other codes.
	int scode = 0;
	double timeshift = p->txtime  -  0.0;
	dprintf("Timeshift: %.9f\n", timeshift);

	double chipshift = timeshift *bw;
	dprintf("Chipshift: %.9f\n", chipshift);
	
//	int chiprem = ((int) rint(chipshift)) % spreading;

	double qt = floor(chipshift/(double)(spreading));
    double remainder = chipshift - ((double)spreading)*qt;
	int chiprem = ((int) rint(remainder));
	dprintf("Right Shift: %d Spreading: %d\n", chiprem, spreading);


	scode = rightshift(p->codeword, chiprem);




	dprintf("Line: ");
	for(int count =(int)spreading -1 ; count >= 0; count--)
		dprintf("%d ", line[count]);

	dprintf("\n New line: ");
	for(int count =(int)spreading -1 ; count >= 0; count--)
	{
	//	dprintf("%.15f\n", pow(2, count));

		if((scode & (int)pow(2, count)) != 0)
			line[count]--;
		if(line[count] < 0)
		{
			dprintf("Error: Removal of codes error\n");
			exit(-1);
		}
		::dprintf("%d ", line[count]);
	}
	numcodesonline--;
	::dprintf("\n");

/*
	if(p->id == 99999)
		sched->halt = 1;
*/

	delete(p);
}

Queue *Line::queue()
{
	return q;
}

void Line::dumpstats()
{

	printf("Packets lost due to interference: %d\n", drops);
	printf("Packets received correctly: %d\n", receives);
	printf("Fraction of packets that were successfully received: %.6f\n", (double)receives/(double)(drops + receives));
	printf("Network throughput: %.6f Mb/s\n", (double)bytes *8.0/(sched->currenttime) * 1e-6);
	//printf("%d %d %.6f %.6f ", drops, receives, (double)receives/(double)(drops + receives), (double)bytes *8.0/(sched->currenttime) * 1e-6);
}


void Line::check(Packet *p)
{

//Check interference for every other packet.

	dprintf("New check: Packet od: %d\n", p->id);
//TBD: Check on packets on the channel being listened to.


//Calculate detected signal
//Add this code to line


	int scode = 0;
	double timeshift = p->txtime  -  0.0;
	dprintf("Timeshift  %.12f\n", timeshift); 
	
	double chipshift = timeshift *bw;
	dprintf("Chip shift : %.12f\n", chipshift); 
/*
	int chiprem = ((int) rint(chipshift)) % spreading;
*/	
	double qt = floor(chipshift/(double)(spreading));
	dprintf("qt: %.12f\n", qt); 
    double remainder = chipshift - ((double)spreading)*qt;
	dprintf("remainder: %.12f\n", remainder); 
	int chiprem = ((int) rint(remainder));
	dprintf("Right Shift: %d Spreading: %0.6f\n", chiprem, spreading);


	scode = rightshift(p->codeword, chiprem);
	dprintf("Adding shifted code: %d\n", scode);

	dprintf("Line: ");
	for(int count =(int)spreading -1; count >=0; count--)
	{
		if((scode & (int)pow(2, count)) != 0)
			line[count]++;
		::dprintf("%d ", line[count]);
	}
	numcodesonline++;
	::dprintf("\n");

	dprintf("Num codes on line: %d\n", numcodesonline);


	Packet *qp = q->head;
	while(qp->next != NULL)
	{

		if ((qp == NULL) || (qp->drop == 1))
		{
			qp=qp->next;
			continue;
		}
		
		

	
//First substract this code from the line
//Shift it 
		int scode = 0;
		double timeshift = qp->txtime  -  0.0;
	

		double chipshift = timeshift *bw;
	//	int chiprem = ((int) rint(chipshift)) % spreading;

	double qt = floor(chipshift/(double)(spreading));
    double remainder = chipshift - ((double)spreading)*qt;
	int chiprem = ((int) rint(remainder));



		dprintf("Right Shift: %d Spreading: %0.6f\n", chiprem, spreading);
		scode = rightshift(qp->codeword, chiprem);
	
		dprintf("Subtracting shifted code: %d\n", scode);

//Subtract it
		int detected =0;
		int det = 0;
		dprintf("Line: ");
		for(int count3 =(int)spreading  - 1; count3 >=0; count3--)
		{
			if((scode & (int)pow(2, count3)) != 0)
				det = line[count3] - 1;
			else
				det = line[count3];

			if(det < 0)
			{
				dprintf("Error: Detecting\n");
				exit(-1);
			}
			::dprintf("%d ", line[count3]);
			if (det !=0)
				detected = detected | (int)pow(2, count3);

		}
		::dprintf("\n");
		dprintf("Detected: %d\n", detected);	



//Then check for interference
//If it interferes with this channel, drop this channel
		if ( scode == ( detected &  scode))
		{
			dprintf("Need to drop packet id: %u\n", qp->id);
			dprintf("Interference module::check(): Dropped packet\n");
			fflush(stdout);
			qp->drop = 1;
			dprintf("Interference module::check(): Dropping packet: %d\n", drops);
		}
		else
		{
				dprintf("Interference module::check(): No interference: %d\n", qp->id);
				dprintf("Interference module::check(): Recvd packet: %.6f\n", receives);
		}	
	
		fflush(stdout);
		qp=qp->next;
	}

	return;
}


int Line:: rightshift( int code, int shift)
{
    int shifted =code;
	for(int count=0; count< shift; count++)
	{
		int lsb = shifted & 1;
//      dprintf("lsb: %d\n", lsb);
		if (lsb != 1)
			shifted = shifted >> 1;
		else
			shifted = (shifted >> 1) | (int)(pow(2, spreading - 1)) ;
        dprintf("Right shifted: %d\n", shifted);
	}
	return shifted;
}

int Line:: leftshift( int code, int shift)
{
    int shifted =code;
	for(int count=0; count< shift; count++)
	{
		int msb = shifted & (int)pow(2, spreading -1);
//      dprintf("lsb: %d\n", lsb);
		if (msb ==0)
			shifted = shifted << 1;
		else
			shifted = (shifted << 1) | 1 ;

		shifted = shifted & ((int)pow(2, spreading) - 1);
        dprintf("Left shifted: %d\n", shifted);
	}
	return shifted;
}
