/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 July 2003
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <ipport.h>
#include <sense.h>
#include <line.h>

#include <debug.h>

Scheduler * sched;
int numNodes;
Sense *s;
Line *l;
IPPort *ip;
char inputfile[20];
char outputfile[20];

char codefile[20];;
double load;

int alg;
int spreading;


int main(int argc,  char **argv)
{

	
	if(argc != 12)
	{
		printf("Usage: linesim <Chipping rate in Mb/s> <Codeset length> <Weight of codeset> <Kappa>  <Sensing window in bits> <Trace file specifier (trace.$specifier.in)>  <Media access protocol/algorithm> <othreshold> <mthreshold> <Backoff duration in us> <Seed>\n");
		printf("Algorithm: 0-Aloha 1-Selfish 2-Threshold 3-Code estimation\n");
		exit(0);
	}



	double bw = 1E6 *(double)atoi(argv[1]); //Mbps
	spreading = atoi(argv[2]); //Mbps
	int weight = atoi(argv[3]);
	int kappa = atoi(argv[4]);
	int window = atoi(argv[5]);
	int inter = atoi(argv[6]);
	alg = atoi(argv[7]);
	int othreshold = atoi(argv[8]);
	int mthreshold = atoi(argv[9]);
	double backoff  = atof(argv[10]) *1e-6;
	int seed = atoi(argv[11]);

	srandom(seed);

	sprintf(codefile, "codes.%d.%d.%d.txt", spreading, weight, kappa);


	sprintf(inputfile, "trace.%d.in", inter);
	sprintf(outputfile, "trace.%d.out", inter);

	sched = new Scheduler();
	dprintf("res.cc: main(): Arguments: Spreading: %d\n", spreading);


	dprintf("res.cc: main(): Arguments: BW: %0.6f bps\n", bw);

	dprintf("res.cc: main(): Created opports\n");

	ip  = new IPPort(spreading);   
	s = new Sense(bw, spreading, window, othreshold, mthreshold, backoff);
	l = new Line(bw, spreading);

//The input interface consists of an IPPort which accepts the incoming packets,
//a router process which searches for the packet in the cache and the cache itself.
	dprintf("res.cc: main(): Created ipports, scheduler\n");
	dprintf("res.cc: main(): Entering scheduler loop\n");
	ip->read();


	printf("Simulation in progress.....\n\n");
	while (1)
		if(sched->run() == 0)
			break;
	printf("Simulation paramters:\n");
	printf("-------------------\n");
	printf("Code: (%d, %d, %d)\n", spreading, weight, kappa);
	printf("Sensing window (bits): %d\n", window);
	printf("Trace file: trace.%d.in\n", inter);
	printf("Media access protocol/algorithm: %d\n", alg);
	printf("Thresholds: %d %d\n", othreshold, mthreshold);
	printf("Backoff time (s): %.6f\n", backoff);
	printf("Random seed: %d\n", seed);
	printf("-------------------\n\n");


	printf("Simulation results:\n");
	printf("-------------------\n");
	s->dumpstats();
	l->dumpstats();
	printf("-------------------\n\n");

	dprintf("res.cc: main(): No more events to schedule\n");
	

	printf("\n");
	printf("Time simulated: %.9f s\n", sched->currenttime - sched->start);	
	
	delete ip;
exit(0);
}
